/*
 +========================================================================+
 |         (c) 2020 - PROPERTY OF ECOLE POLYTECHNIQUE - LGPL 3.0          |
 |________________________________________________________________________|
 |   '&`   |                                                              |
 |    #    |   FILE       : head.cpp                                      |
 |    #    |   VERSION    : 0.1.0                                         |
 |   _#_   |   AUTHOR(S)  : Matthieu Aussal                               |
 |  ( # )  |   CREATION   : 01.04.2020                                    |
 |  / 0 \  |   LAST MODIF : 31.10.2020                                    |
 | ( === ) |   SYNOPSIS   : Acoustical scattering using BEM with Neumann  |
 |  `---'  |                boundary condition                            |
 +========================================================================+
 */

#include <castor/matrix.hpp>
#include <castor/smatrix.hpp>
#include <castor/hmatrix.hpp>
#include <castor/linalg.hpp>
#include <castor/graphics.hpp>
#include "bembuilder.hpp"

using namespace castor;

int main (int argc, char* argv[])
{
    // Load meshes
    matrix<double> Svtx;
    matrix<size_t> Stri;
    std::tie(Stri,Svtx) = triread("/Users/aussal/Git/leprojetcastor/bembuilder/data/","Head03_04kHz.ply");
    
    // Graphical representation
    figure fig;
    trimesh(fig,Stri,Svtx);
    
    // Parameters
    matrix<double> U = {0,0,-1};
    double f = 4000;
    double k = 2*M_PI*f/340;
    float tol = 1e-3;
    
    // FEM
    tic();
    femdata<double> v(Stri,Svtx,lagrangeP1,3);
    femdata<double> u(Stri,Svtx,lagrangeP1,3);
    toc();
    
    // Identity matrix
    tic();
    auto Id = mass<std::complex<double>>(v);
    toc();

    // Left hand side : [-H]
    tic();
    auto LHSfct = [&v,&u,&k](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return -hypersingular<std::complex<double>>(v,u,k,Ix,Iy);
    };
    hmatrix<std::complex<double>> LHS(v.dof(),u.dof(),tol,LHSfct);
    toc();
    disp(LHS);

    // Right hand side : - \int_sigma_x v(x) PW(x,u) dx
    auto B = - rightHandSide<std::complex<double>>(v,dnPWsource,U,k);

    // Solve -H = -dnP0
    hmatrix<std::complex<double>> Lh,Uh;
    tic();
    std::tie(Lh,Uh) = lu(LHS,1e-1);
    toc();
    disp(Lh);
    disp(Uh);
    auto mu = gmres(LHS,B,tol,100,Lh,Uh);
    
    // Boundary radiation
    tic();
    auto Dbndfct = [&v,&u,&k,&Id](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return 0.5*eval(Id(Ix,Iy)) + doubleLayer<std::complex<double>>(v,u,k,Ix,Iy);
    };
    matrix<std::complex<double>> Pbnd = zeros(size(v.dof(),1),size(mu,2));
    hmatrix<std::complex<double>> Dbnd(v.dof(),u.dof(),tol,Dbndfct,mu,Pbnd);
    toc();
    Pbnd = - gmres(Id,Pbnd,tol,100) + planeWave(v.dof(),U,k);
    
    // Graphical representation
    figure fig2;
    trimesh(fig2,Stri,Svtx,real(Pbnd));
    
    // Export in .vtk file
    triwrite("/Users/aussal/Desktop/resultats/","head.vtk",Stri,Svtx,real(Pbnd));
    
    // Plot
    drawnow(fig);
    disp("done !");
    return 0;
}
